import logging
import os
import pathlib
import queue
import threading
import traceback
from time import sleep
from types import UnionType
from typing import Any, Callable, Generic, Type, TypeAlias, TypeVar, cast


class CallableWrapper:
    def __init__(self, func: Callable[[], None]):
        self._func = func
        self.ready = False

    def __call__(self):
        try:
            self._func()
        finally:
            self.ready = True

    def wait(self, sleep_seconds: float = 0.001):
        while not self.ready:
            sleep(sleep_seconds)


_thread_sentinel = threading.Thread()


class ThreadRouter:
    def __init__(self, target_thread: threading.Thread = _thread_sentinel):
        self._target_thread = (
            target_thread if target_thread is not _thread_sentinel else threading.current_thread()
        )
        self._queue = queue.Queue[Callable[[], None]]()

    def exec(self, action: Callable[[], None]):
        if threading.current_thread() == self._target_thread:
            action()
        else:
            wrap = CallableWrapper(action)
            self._queue.put(wrap)
            wrap.wait()

    def process(self):
        if threading.current_thread() != self._target_thread:
            raise RuntimeError("wrong thread")
        while True:
            try:
                item = self._queue.get(False)
            except queue.Empty:
                return
            try:
                item()
            except Exception:
                logging.error(traceback.format_exc())


def get_data_dir() -> pathlib.Path:
    home = pathlib.Path.home()
    if "/" in str(home):
        return home.joinpath(".local", "share", "kvix")
    elif "\\" in str(home):
        appdata = os.getenv("APPDATA")
        if appdata:
            return pathlib.Path(appdata, "kvix")
        else:
            return home.joinpath("AppData", "kvix")
    else:
        raise RuntimeError("unexpected platform separator derived from HOME")


def get_config_dir() -> pathlib.Path:
    home = pathlib.Path.home()
    if "/" in str(home):
        return home.joinpath(".config", "kvix")
    elif "\\" in str(home):
        appdata = os.getenv("APPDATA")
        if appdata:
            return pathlib.Path(appdata, "kvix")
        else:
            return home.joinpath("AppData", "roaming", "kvix")
    else:
        raise RuntimeError("unexpected platform separator derived from HOME")


def get_cache_dir() -> pathlib.Path:
    home = pathlib.Path.home()
    if "/" in str(home):
        return home.joinpath(".cache", "kvix")
    elif "\\" in str(home):
        appdata = os.getenv("LOCALAPPDATA")
        if appdata:
            return pathlib.Path(appdata, "kvix")
        else:
            return home.joinpath("AppData", "kvix")
    else:
        raise RuntimeError("unexpected platform separator derived from HOME")


key_mappings: list[dict[str, str]] = []
key_sets: list[str] = [
    "abcdefghijklmnopqrstuvwxyz",
    "фисвуапршолдьтщзйкыегмцчня",
]
for key_set in key_sets:
    for other_key_set in key_sets:
        if key_sets is other_key_set:
            continue
        key_mappings.append(dict(zip(list(key_set), list(other_key_set))))


def query_match(query: str, *contents: str) -> bool:
    if not query:
        return True

    def normalize(x: str):
        return x.lower().replace("ё", "е")

    for word in query.split():
        if not word:
            continue
        word = normalize(word)
        for item in contents:
            if word in normalize(item.lower()):
                return True
    return False


class Sentinel:
    pass


_sentinel = Sentinel()

T = TypeVar("T")


class Propty(Generic[T]):
    """
    Usage variants:
    x = Propty(str) # just wrapper around self._x
    # like previous, with additional customizations
    x = Propty(dict,default_supplier=lambda: {'a':1}, private_name='_private_x_value',
            on_change='_on_change_x')
    # alternative to x = property(fget=_get_x, fset=_set_x)
    x = Propty(str, getter = _get_x, setter = _set_x)
    # just like previous but supports override of getter and setter in child classes
    x = Propty(str, getter = '_get_x', setter = '_set_x')

    """

    @staticmethod
    def create_prototype(**fix_kwargs: Any):
        def result(*args: Any, **kwargs: Any):
            kwargs.update(fix_kwargs)
            return Propty(*args, **kwargs)

        return result

    _type_default_value = cast(Type[T], None)
    _defaut_value_defalt_value = cast(T, _sentinel)
    _default_supplier_default_value = lambda: cast(T, None)

    def __init__(
        self,
        type: Type[T] = _type_default_value,
        # default value supply
        default_value: T = _defaut_value_defalt_value,
        # type default constructor by default
        default_supplier: Callable[[], T] = _default_supplier_default_value,
        value_predicate: Callable[[T], bool] = lambda x: x and True or False,
        # change notification
        on_change: str | bool | Callable[[Any, T], None] = False,
        comparator: Callable[[T, T], bool] = lambda a, b: True,
        #
        type_check: bool = False,
        writeable: bool = True,
        required: bool = False,
        # storage
        private_name: str = "",
        getter: str | Callable[[Any], T] | None = None,
        setter: str | Callable[[Any, T], None] | None = None,
    ):
        self._name = ""
        self._type = type
        if not writeable and setter:
            raise RuntimeError("Propty: not writeable and setter")
        if required and bool(default_supplier):
            raise RuntimeError("Propty: required and default_supplier")

        # default value supplier
        def get_default_supplier():
            if default_supplier != Propty._default_supplier_default_value:
                return default_supplier
            if default_value is not _sentinel:
                return lambda: default_value
            if type:
                return type
            return lambda: None

        self._default_supplier = get_default_supplier()

        self._value_predicate = value_predicate
        # change notificate
        self._on_change = on_change
        self._comparator = comparator
        # self._require_listener_exists = self._on_change
        #
        if type_check and not type:
            raise RuntimeError("Propty: type_check and not type")
        self._type_check = type_check
        self._writeable = writeable
        self._required = required
        # storage
        self._private_name = private_name
        self._getter = getter or self._builtin_getter
        self._setter = setter or self._builtin_setter

    def __set_name__(self, owner: Any, name: str):
        self._name = name
        if self._on_change is True:
            self._on_change = "_on_change_" + name
        if not self._private_name:
            self._private_name = "_" + name

    def __set__(self, obj: Any, value: T):
        if not self._writeable:
            raise AttributeError("property for " + self._name + " is not writeable")
        self._assert_type(value)
        self._maybe_notify_change_listeners(obj, value)
        self._invoke_method(self._setter, obj, value)

    def _assert_type(self, value: Any):
        if not self._type_check:
            return
        assert self._type
        if not isinstance(value, self._type):
            raise AttributeError("property " + self._name + ": unexpected type")

    def _maybe_notify_change_listeners(self, obj: Any, value: T) -> None:
        "invoke listeners, if they are configured and value really changed"
        # check if there are change listeners
        if not self._on_change:
            return
        # check if value has really changed
        self._assert_obj(obj)
        old_value: T = self.__get__(obj)
        if self._comparator(old_value, value):
            return
        # invoke listener
        self._invoke_method(self._on_change, obj, value)

    def _invoke_method(self, method: Any, obj: Any, *args: Any, **kwargs: Any):
        if isinstance(method, str) and method:
            self._assert_obj(obj)
            if not hasattr(obj, method):
                raise RuntimeError("method " + method + " not found")
            func = cast(Callable[[T], None], getattr(obj, method))
            if not callable(func):
                raise RuntimeError("attr " + method + " expected to be callable")
            return func(*args, **kwargs)
        elif callable(method):
            self._assert_obj(obj)
            return method(obj, *args, **kwargs)
        else:
            raise AssertionError("callable or non-empty string expected")

    def _builtin_setter(self, obj: Any, value: T) -> None:
        setattr(obj, self._private_name, value)

    def __get__(self, obj: Any, objtype: Any = None) -> T:
        result = self._invoke_method(self._getter, obj)
        self._assert_type(result)
        if not self._value_predicate(cast(T, result)):
            result = self._default_supplier()
            # todo maybe self._assert_type(result)
            self._invoke_method(self._setter, obj, result)
        return cast(T, result)

    def _assert_obj(self, obj: Any) -> Any:
        if not obj:
            raise AttributeError("Propty is for instances only")
        return obj

    def _builtin_getter(self, obj: Any) -> T:
        self._assert_obj(obj)
        if not hasattr(obj, self._private_name):
            # return self._default_supplier()
            return cast(T, None)
        return getattr(obj, self._private_name)


TypeKey = TypeVar("TypeKey")
TypeValue = TypeVar("TypeValue")


def ensure_key(
    dest: dict[TypeKey, TypeValue],
    key: TypeKey,
    supplier: Callable[[], TypeValue],
) -> TypeValue:
    if key in dest:
        return dest[key]
    else:
        value = supplier()
        dest[key] = value
        return value


ClassInfo: TypeAlias = type | UnionType | tuple["ClassInfo", ...]


def ensure_type(value: Any | None, type_or_tuple: ClassInfo):
    if not isinstance(value, type_or_tuple):
        raise RuntimeError("value expected to be %s" % type_or_tuple)
    return value


def apply_template(template: str, **values: str) -> str:
    result = str(template)
    for key, value in values.items():
        result = result.replace("{{" + str(key) + "}}", str(value))
    return result



from __future__ import annotations

import tkinter as tk
from tkinter import (
    Misc as _TkMisc,
    Tk as _Tk,
    Listbox as _TkListbox,
    Entry as _TkEntry,
    Button as _TkButton,
    Scrollbar as _TkScrollbar,
)
from tkinter.ttk import Style as _TkStyle
from typing import Any as _Any, cast as _cast, Generator as _Generator
from .util import find_all_children

from kvix.l10n import _

copy_text = _("Copy").setup(ru_RU="Копировать", de_DE="Kopieren")
cut_text = _("Cut").setup(ru_RU="Вырезать", de_DE="Ausschneiden")
del_text = _("Delete").setup(ru_RU="Удалить", de_DE="Entfernen")
paste_text = _("Paste").setup(ru_RU="Вставить", de_DE="Einfügen")

_style_singleton: BaseStyle = None


def _apply_widget_style(widget: _TkMisc):
    if _style_singleton:
        _style_singleton.configure_widgets(widget)


class Listbox(_TkListbox):
    def __init__(self, master=None, cnf={}, **kw):
        _TkListbox.__init__(self, master, cnf, **kw)
        _apply_widget_style(self)


class Menu(tk.Menu):
    def __init__(
        self,
        master: tk.Tk | None = None,
        cnf: dict[str, _Any] = {},
        **kw: _Any,
    ):
        tk.Menu.__init__(self, master, cnf, **kw)
        _apply_widget_style(self)


class Entry(_TkEntry):
    def __init__(
        self,
        master: tk.Tk | None = None,
        cnf: dict[str, _Any] = {},
        **kw: _Any,
    ):
        _TkEntry.__init__(self, master, cnf, **kw)
        _apply_widget_style(self)
        self._init_menu()

    def _init_menu(self):
        self._popup_menu = Menu(self, tearoff=False)
        self._popup_menu.add_command(
            label=str(copy_text),
            command=lambda: self.event_generate("<<Copy>>"),
        )
        self._popup_menu.add_command(
            label=str(cut_text), command=lambda: self.event_generate("<<Cut>>")
        )
        self._popup_menu.add_command(label=str(del_text), command=self._delete)
        self._popup_menu.add_separator()
        self._popup_menu.add_command(
            label=str(paste_text),
            command=lambda: self.event_generate("<<Paste>>"),
        )
        self.bind(
            "<Button-3>",
            lambda event: self._popup_menu.tk_popup(event.x_root, event.y_root),
        )

    def _delete(self):
        if not self.selection_present():
            return
        self.delete(self.index("sel.first"), self.index("sel.last"))


class Button(_TkButton):
    def __init__(self, master=None, cnf={}, **kw):
        _TkButton.__init__(self, master, cnf, **kw)
        _apply_widget_style(self)


class Scrollbar(_TkScrollbar):
    def __init__(self, master=None, cnf={}, **kw):
        _TkScrollbar.__init__(self, master, cnf, **kw)
        _apply_widget_style(self)


class Theme:
    def __init__(
        self,
        frame_back_color: str = "",
        button_back_color: str = "",
        edit_back_color: str = "",
        back_color: str = "",
        button_text_color: str = "",
        edit_text_color: str = "",
        text_color: str = "",
        select_back_color: str = "",
    ):
        self.frame_back_color = frame_back_color or back_color
        self.button_back_color = button_back_color or frame_back_color or back_color
        self.edit_back_color = edit_back_color or back_color

        self.button_text_color = button_text_color or edit_text_color or text_color
        self.edit_text_color = edit_text_color or button_text_color or text_color
        self.select_back_color = select_back_color

    def create_settings(self) -> dict[str, _Any]:
        return {
            ".": {
                "configure": {
                    "background": self.frame_back_color,
                    "foreground": self.button_text_color,
                    "insertbackground": self.edit_text_color,  # insert blinking cursor color
                    "fieldbackground": self.edit_back_color,
                    "selectbackground": self.select_back_color,
                    "selectforeground": self.edit_text_color,
                    "highlightthickness": "0"
                    # 'highlightcolor': self.highlight_color,
                }
            }
        }

    def configure_widget(self, widget: _TkMisc):
        if "Listbox" == widget.winfo_class():
            _cast(_TkListbox, widget).configure(
                background=self.edit_back_color,
                foreground=self.edit_text_color,
                highlightthickness="0",
            )
        elif "Entry" == widget.winfo_class():
            _cast(_TkEntry, widget).configure(
                background=self.edit_back_color,
                foreground=self.edit_text_color,
                insertbackground=self.edit_text_color,
                highlightthickness="0",
                selectbackground=self.select_back_color,
                selectforeground=self.edit_text_color,
            )
        elif "Button" == widget.winfo_class():
            _cast(_TkButton, widget).configure(
                bg=self.button_back_color, fg=self.button_text_color
            )
        elif "Scrollbar" == widget.winfo_class():
            _cast(_TkScrollbar, widget).configure(background=self.button_back_color)
        elif "Menu" == widget.winfo_class():
            _cast(tk.Menu, widget).configure(
                background=self.button_back_color,
                foreground=self.button_text_color,
                activebackground=self.select_back_color,
                activeforeground=self.button_text_color,
            )


class BaseStyle:
    def __init__(self):
        global _style_singleton
        if _style_singleton:
            raise Exception("kvix.ui.tk.plus:Style should be singleton")
        _style_singleton = self
        self._themes: dict[str, Theme] = {}
        self._current_theme_name: str = ""

    def theme_register(self, theme_name: str, theme: Theme) -> None:
        if theme_name in self._themes:
            raise Exception("duplicate theme: " + theme_name)
        self._themes[theme_name] = theme

    def theme_use(self, theme_name: str, root_widget: _TkMisc):
        if not theme_name:
            return self._current_theme_name
        if not theme_name in self._themes:
            return
        self._current_theme_name = theme_name
        self.configure_widgets(root_widget)
        return theme_name

    def configure_widgets(self, root: _TkMisc):
        if not self._current_theme_name:
            return
        current_theme = self._themes[self._current_theme_name]
        if not current_theme:
            return
        for widget in find_all_children(root):
            current_theme.configure_widget(widget)


class Style(BaseStyle, _TkStyle):
    def __init__(self, master: _Tk | None = None):
        _TkStyle.__init__(self, master)
        BaseStyle.__init__(self)
        self.theme_register(
            "darcula",
            "clam",
            Theme(
                frame_back_color="#3c3f41",
                button_back_color="#333333",
                edit_back_color="#282829",
                text_color="#adadad",
                select_back_color="#173a7b",
            ),
        )

    def theme_register(self, theme_name: str, parent: str | None, theme: Theme):
        BaseStyle.theme_register(self, theme_name, theme)
        _TkStyle.theme_create(self, theme_name, parent, theme.create_settings())

    def theme_use(self, themename: None = None) -> str | None:
        if themename is None:
            return _TkStyle.theme_use(self)
        _TkStyle.theme_use(self, themename)
        if self.master:
            BaseStyle.theme_use(self, str(themename), self.master)



import importlib.util
import os
import sys
from pathlib import Path
from types import ModuleType

from kvix.util import get_data_dir

from kvix import Context
from kvix.l10n import _
from kvix.plugin.impl import FromModules

plugin_dir_config_item_text = _("Plugin directory").setup(
    ru_RU="Папка для плагинов", de_DE="Plugin Ordner"
)


def scan_python_packages(root_dir: Path | str):
    if not Path(root_dir).is_dir():
        return
    for item in os.scandir(Path(root_dir)):
        if item.is_file() and item.name.endswith(".py"):
            yield (item.name, item.path)


def load_module_from_path(name: str, path: Path | str) -> ModuleType:
    spec = importlib.util.spec_from_file_location(name, path)
    if not spec:
        raise RuntimeError("unloadable module")
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    spec.loader.exec_module(module)
    return module


def load_plugin_modules(root_dir: Path | str) -> list[ModuleType]:
    result = []
    for name, path in scan_python_packages(root_dir):
        try:
            result.append(load_module_from_path(__name__ + "." + name, path))
        except Exception as e:
            print(e)
    return result


def get_plugin_dir(context: Context):
    conf_item = context.conf.item("plugin_dir")
    conf_item.setup(
        title=str(plugin_dir_config_item_text),
        default=str(get_data_dir().joinpath("plugins")),
    )
    result = Path(conf_item.read())

    if not os.path.exists(result):
        os.makedirs(result)
    return result


class Plugin(FromModules):
    def __init__(self, context: Context):
        FromModules.__init__(self, context, *load_plugin_modules(get_plugin_dir(context)))



from __future__ import annotations

from kvix.util import Propty
import os.path
import pathlib
from typing import Any

import yaml


class Stor:
    data = Propty(Any, default_supplier=dict)

    def load(self) -> None:
        ...

    def save(self) -> None:
        ...


class YamlFile(Stor):
    def __init__(self, file_path: pathlib.Path):
        self._file_path = file_path

    def load(self) -> None:
        if not os.path.exists(self._file_path):
            self.data = None
            return
        with open(self._file_path, "r") as stream:
            self.data = yaml.safe_load(stream)

    def save(self) -> None:
        dir_path = os.path.dirname(self._file_path)
        if not os.path.exists(dir_path):
            os.makedirs(
                dir_path,
            )
        with open(self._file_path, "w") as out:
            yaml.dump(self.data, out, default_flow_style=False, allow_unicode=True)



from __future__ import annotations
from types import ModuleType
from kvix.impl import BasePlugin, FromModule
from kvix import Action, ActionType, Context


class Compound(BasePlugin):
    def __init__(self, context: Context, *wrap: kvix.Plugin):
        BasePlugin.__init__(self, context)
        self.wrap = wrap

    def get_action_types(self) -> list[ActionType]:
        return [action_type for plugin in self.wrap for action_type in plugin.get_action_types()]

    def get_actions(self) -> list[Action]:
        return [action for plugin in self.wrap for action in plugin.get_actions()]


class FromModules(Compound):
    def __init__(self, context: Context, *modules: ModuleType):
        Compound.__init__(self, context, *[FromModule(context, module) for module in modules])



from __future__ import annotations

from typing import Any, Callable

from kvix.util import Propty
from kvix.conf import Conf


class ItemAlt:
    def execute(self) -> None:
        raise NotImplementedError()


class Item:
    priority = Propty(int)
    alts = Propty(list[ItemAlt], writeable=False)


class ItemSource:
    def search(self, query: str) -> list[Item]:
        raise NotImplementedError()


class Context:
    conf = Propty(Conf, writeable=False)
    ui: Propty[Ui] = Propty(writeable=False)
    action_registry: Propty[ActionRegistry] = Propty(writeable=False)

    def quit(self) -> None:
        raise NotImplementedError()


class ActionType:
    context = Propty(type=Context)
    id = Propty(type=str)
    title = Propty(type=str)

    def action_from_config(self, value: Any) -> Action:
        raise NotImplementedError()

    def create_editor(self, builder: DialogBuilder) -> None:
        raise NotImplementedError()


class Action:
    action_type = Propty(ActionType)
    title = Propty(str)
    description = Propty(str)

    def search(self, query: str) -> list[Item]:
        raise NotImplementedError()

    def to_config(self) -> dict[str, Any]:
        raise NotImplementedError()


class ActionRegistry(ItemSource):
    action_types = Propty(dict[str, ActionType], writeable=False)
    actions = Propty(list[Action], writeable=False)

    def load(self) -> None:
        raise NotImplementedError()

    def save(self) -> None:
        raise NotImplementedError()

    def add_action_type(self, action_type: ActionType) -> None:
        raise NotImplementedError()

    def action_from_config(self, value: Any) -> Action:
        raise NotImplementedError()

    def search(self, query: str) -> list[Item]:
        raise NotImplementedError()


class Ui:
    def on_ready(self, f: Callable[[], None]) -> None:
        raise NotImplementedError()

    def run(self) -> None:
        raise NotImplementedError()

    def selector(self) -> Selector:
        raise NotImplementedError()

    def dialog(self, create_dialog: Callable[[DialogBuilder], None]) -> Dialog:
        raise NotImplementedError()

    def destroy(self) -> None:
        raise NotImplementedError()

    def copy_to_clipboard(self, data: bytes) -> None:
        raise NotImplementedError()

    def paste_from_clipboard(self) -> bytes:
        raise NotImplementedError()

    def hide(self) -> None:
        "hide and release focus"
        raise NotImplementedError()


class Window:
    title = Propty(str, default_value="kvix")

    def go(self) -> None:  # todo rename to activate
        raise NotImplementedError()

    def destroy(self) -> None:
        raise NotImplementedError()


class Selector(Window):
    item_source = Propty(ItemSource)

    def hide(self) -> None:
        raise NotImplementedError()


class Dialog(Window):
    value = Propty()
    on_ok = Propty(Callable[[], None], default_value=lambda: None)
    on_cancel = Propty(Callable[[], None], default_value=lambda: None)
    auto_destroy = Propty(bool, default_value=True)  # todo remove?


class DialogWidget:
    def get_value(self) -> str:
        raise NotImplementedError()

    def set_value(self, value: str) -> None:
        raise NotImplementedError()


class DialogEntry(DialogWidget):
    pass


class DialogBuilder:
    def __init__(self):
        self._on_load: list[Callable[[Any | None], None]] = []
        self._on_save: list[Callable[[Any | None], Any]] = []
        self._widgets: dict[str, DialogWidget] = {}

    def create_entry(self, id: str, title: str) -> DialogEntry:
        raise NotImplementedError()

    def widget(self, id: str) -> DialogWidget:
        return self._widgets[id]

    def _add_widget(self, id: str, widget: DialogWidget) -> DialogWidget:
        self._widgets[id] = widget
        return widget

    def on_load(self, func: Callable[[Any | None], None]):
        self._on_load.append(func)

    def load(self, value: Any | None):
        for func in self._on_load:
            func(value)

    def on_save(self, func: Callable[[Any | None], Any]):
        self._on_save.append(func)

    def save(self, value: Any) -> Any | None:
        for func in self._on_save:
            value = func(value)
        return value


class Plugin:
    def __init__(self, context: Context):
        ...

    def get_action_types(self) -> list[ActionType]:
        ...

    def get_actions(self) -> list[Action]:
        ...
